@@ -1,5 +1,5 @@
 /*
- * uicode_tool - Manipulates Intel(R) IA32/x86_64 processor microcode bundles
+ * iucode_tool - Manipulates Intel(R) IA32/x86_64 processor microcode bundles
  *
  * Copyright (c) 2010-2013 Henrique de Moraes Holschuh <hmh@hmh.eng.br>
  *               2000 Simon Trimmer, Tigran Aivazian
@@ -867,7 +867,11 @@ static int load_intel_microcode(const char * path,
 			}
 		}
 
-		fd = open(fn, O_RDONLY);
+		if (dir) {
+			fd = openat(dirfd(dir), dentry->d_name, O_RDONLY);
+		} else {
+			fd = open(fn, O_RDONLY);
+		}
 		if (unlikely(fd == -1)) {
 			err = errno;
 			print_err("%s: cannot open: %s", fn, strerror(err));
@@ -1007,10 +1011,12 @@ static int upload_intel_microcodes(const char * const devname,
 	if (fstat(fd, &stat) == -1) {
 		err = errno;
 		print_err("%s: cannot stat: %s", devname, strerror(err));
+		close(fd);
 		return err;
 	}
 	if (!S_ISCHR(stat.st_mode)) {
 		print_err("%s: not a character device", devname);
+		close(fd);
 		return EINVAL;
 	}
 
@@ -1161,12 +1167,16 @@ static int __write_cpio_trailer(int fd, size_t size)
 
 /**
  * write_intel_microcodes() - writes microcodes to bin file
+ * @dirfd:		directory where to create file
  * @filename:		file to write to
  * @ft:			file type
  * @uc_write_list:	uclist with the microcodes to write
  *
  * returns 0 if successful, or errno() if a problem happens
  *
+ * @dirfd should be either an open directory for openat(), or
+ * the special value AT_FDCWD.
+ *
  * @ft should be zero for binary format, 1 for linux early
  * initramfs cpio format.
  *
@@ -1180,7 +1190,8 @@ static int __write_cpio_trailer(int fd, size_t size)
  * newer.  This file should be prepended to the compressed initramfs
  * image.
  */
-static int write_intel_microcodes(const char * const filename, int ft,
+static int write_intel_microcodes(const int dirfd,
+				  const char * const filename, int ft,
 				  struct intel_uclist_entry *uc_write_list)
 {
 	int fd;
@@ -1195,7 +1206,7 @@ static int write_intel_microcodes(const char * const filename, int ft,
 
 	/* Unlink first */
 	if (unlink_files) {
-		if (unlink(filename) == -1) {
+		if (unlinkat(dirfd, filename, 0) == -1) {
 			if (errno != ENOENT) {
 				err = errno;
 				print_err("%s: cannot unlink: %s", filename,
@@ -1206,8 +1217,9 @@ static int write_intel_microcodes(const char * const filename, int ft,
 			print_msg(3, "unlinked %s", filename);
 		}
 	}
-	fd = open(filename, O_CREAT | O_WRONLY | O_EXCL,
-			    S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
+
+	fd = openat(dirfd, filename, O_CREAT | O_WRONLY | O_EXCL,
+		    S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
 	if (fd == -1) {
 		err = errno;
 		print_err("%s: cannot write to, or create file: %s",
@@ -1658,7 +1670,7 @@ static int do_write_microcode(const char * const filename, int ft)
 		return 0;
 
 	print_msg(1, "Writing selected microcodes to: %s", filename);
-	return write_intel_microcodes(filename, ft, microcodes);
+	return write_intel_microcodes(AT_FDCWD, filename, ft, microcodes);
 }
 
 static int do_upload_microcode(const char * const filename)
@@ -1672,7 +1684,8 @@ static int do_upload_microcode(const char * const filename)
 
 static int do_write_named(const char * const dirname)
 {
-	char fn[PATH_MAX];
+	char fn[35]; /* "s%08X_m%08X_r%08X.fw" */
+	int dirfd;
 	struct stat st;
 
 	struct intel_uclist_entry e, *p;
@@ -1682,16 +1695,25 @@ static int do_write_named(const char * const dirname)
 	if (!microcodes)
 		return 0;
 
+	dirfd = open(dirname, O_RDONLY);
+	if (dirfd == -1) {
+		rc = errno;
+		print_err("%s: cannot open: %s", dirname,
+			  strerror(rc));
+		return rc;
+	}
+
 	/* were we given a directory ? */
-	if (stat(dirname, &st) == -1) {
-		int err = errno;
+	if (fstat(dirfd, &st) == -1) {
+		rc = errno;
 		print_err("%s: cannot stat inode: %s", dirname,
-			  strerror(err));
-		return err;
+			  strerror(rc));
+		goto err_exit;
 	}
 	if (!S_ISDIR(st.st_mode)) {
 		print_err("%s: is not a directory", dirname);
-		return EINVAL;
+		rc = EINVAL;
+		goto err_exit;
 	}
 
 	print_msg(1, "Writing microcode file(s) into %s", dirname);
@@ -1701,18 +1723,12 @@ static int do_write_named(const char * const dirname)
 	rc = 0;
 
 	while (p && !rc) {
-		/* write to file in dirname/ */
-		rc = snprintf(fn, sizeof(fn), "%s/s%08X_m%08X_r%08X.fw", dirname,
+		snprintf(fn, sizeof(fn), "s%08X_m%08X_r%08X.fw",
 			     p->cpuid, p->pf_mask, p->uc_rev);
-		if (rc < 1 || rc >= sizeof(fn)) {
-			print_err("%s: directory name too long, cannot create files inside",
-				  dirname);
-			return EINVAL;
-		}
 
 		memcpy(&e, p, sizeof(e));
 		e.next = NULL;
-		rc = write_intel_microcodes(fn, 0, &e);
+		rc = write_intel_microcodes(dirfd, fn, 0, &e);
 		if (!rc)
 			count++;
 
@@ -1723,12 +1739,16 @@ static int do_write_named(const char * const dirname)
 	else
 		print_msg(1, "no files were written into %s", dirname);
 
+err_exit:
+	close(dirfd);
+
 	return rc;
 }
 
 static int do_write_firmware(const char * const dirname)
 {
-	char fn[PATH_MAX];
+	char fn[35]; /* "%02x-%02x-%02x" */
+	int dirfd;
 	struct stat st;
 
 	struct intel_uclist_entry *samecpuid_list, *p;
@@ -1738,16 +1758,25 @@ static int do_write_firmware(const char * const dirname)
 	if (!microcodes)
 		return 0;
 
+	dirfd = open(dirname, O_RDONLY);
+	if (dirfd == -1) {
+		rc = errno;
+		print_err("%s: cannot open: %s", dirname,
+			  strerror(rc));
+		return rc;
+	}
+
 	/* were we given a directory ? */
-	if (stat(dirname, &st) == -1) {
-		int err = errno;
+	if (fstat(dirfd, &st) == -1) {
+		rc = errno;
 		print_err("%s: cannot stat inode: %s", dirname,
-			  strerror(err));
-		return err;
+			  strerror(rc));
+		goto err_exit;
 	}
 	if (!S_ISDIR(st.st_mode)) {
 		print_err("%s: is not a directory", dirname);
-		return EINVAL;
+		rc = EINVAL;
+		goto err_exit;
 	}
 
 	print_msg(1, "Writing microcode firmware file(s) into %s", dirname);
@@ -1770,8 +1799,10 @@ static int do_write_firmware(const char * const dirname)
 						p->cpuid, p->pf_mask,
 						p->uc_rev, p->uc, p->uc_size,
 						0, &samecpuid_list);
-			if (__uclist_add_print_errors(add_status))
-				return add_status;
+			if (__uclist_add_print_errors(add_status)) {
+				rc = add_status;
+				goto err_exit;
+			}
 
 			p = p->next;
 		}
@@ -1786,15 +1817,10 @@ static int do_write_firmware(const char * const dirname)
 		if (x86_family >= 0x06)
 			x86_model += ((cpuid >> 16) & 0x0f) << 4;
 
-		rc = snprintf(fn, sizeof(fn), "%s/%02x-%02x-%02x", dirname,
+		snprintf(fn, sizeof(fn), "%02x-%02x-%02x",
 			     x86_family, x86_model, x86_mask);
-		if (rc < 1 || rc >= sizeof(fn)) {
-			print_err("%s: directory name too long, cannot create files inside",
-				  dirname);
-			return EINVAL;
-		}
 
-		rc = write_intel_microcodes(fn, 0, samecpuid_list);
+		rc = write_intel_microcodes(dirfd, fn, 0, samecpuid_list);
 
 		free_uclist(samecpuid_list);
 		samecpuid_list = NULL;
@@ -1807,6 +1833,9 @@ static int do_write_firmware(const char * const dirname)
 	else
 		print_msg(1, "no files were written into %s", dirname);
 
+err_exit:
+	close(dirfd);
+
 	return rc;
 }
 
@@ -1921,7 +1950,7 @@ err_out:
 
 /* Command line processing */
 
-const char program_version[] =
+static const char program_version[] =
 	PROGNAME " " VERSION "\n"
 	"Copyright (c) 2010-2013 by Henrique de Moraes Holschuh\n\n"
 
@@ -1932,7 +1961,7 @@ const char program_version[] =
 	"This is free software; see the source for copying conditions.\n"
 	"There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR\n"
 	"A PARTICULAR PURPOSE.";
-const char program_bug_address[] = PACKAGE_BUGREPORT;
+static const char program_bug_address[] = PACKAGE_BUGREPORT;
 
 static char cmdline_doc[] =
 	PROGNAME " - Tool to manipulate Intel IA32/X86_64 microcode bundles\n"
@@ -2399,7 +2428,7 @@ int main(int argc, char *argv[])
 	argp_err_exit_status = EXIT_USAGE;
 	argp_program_version = program_version;
 	argp_program_bug_address = NULL; /* FIXME */
-	argp_parse(&cmdline_argp, argc, argv, ARGP_IN_ORDER, 0, NULL);
+	argp_parse(&cmdline_argp, argc, argv, ARGP_IN_ORDER, NULL, NULL);
 
 	if (!command_line_actions) {
 		print_msg(1, "nothing to do...");
